import os
import sys
import compiler
from parser import ParserError
from bike.parsing.load import load as loadast
from bike.refactor import extractMethod
from bike.refactor.rename import rename
from bike.refactor.extractMethod import coords
from bike.parsing.output import save as saveast
from bike.parsing.utils import fqn_rcar, fqn_rcdr
from bike.parsing import visitor
from bike.parsing.undo import getUndoStack, UndoStackEmptyException
from bike.parsing.fastparserast import Class, Function
from bike.query.common import getScopeForLine
from bike.query.getTypeOf import getTypeOfExpr, UnfoundType
from bike.query.findReferences import findReferences
from bike.query.findDefinition import findAllPossibleDefinitionsByCoords
from bike import log


# Returns a BRMContext object
# wants to indicate progress. The object must be callable and takes
# the following parameters:
#  - action  (string)
#  - filename (string)
def load(path):
    context = init()
    context.load(path)
    return context

def init():
    context = BRMContext_impl()
    return context

# the context object public interface
class BRMContext(object):

    def load(self, path):
        """ import some more code into bicyclerepairman """

    def isLoaded(self, filename):
        """ has the file ever been imported? """

    def save(self):
        """ save the changed files out to disk """

    def setRenameMethodPromptCallback(self, callback):
        """
        sets a callback to ask the user about method refs which brm
        can't deduce the type of. The callback must be callable, and
         take the following parameters:
          - filename
          - linenumber
          - begin column
          - end column
         (begin and end columns enclose the problematic method call)
        """
        
    def getFullyQualifiedNameOfScope(self,filename_path,line):
        """ converts line number into an fqn of the enclosing scope
        (e.g. a.b.foo.TheClass)"""

    def getTypeOfExpression(self, filename_path, line,
                           colbegin, colend):        
        """ returns a type object (which has attribute 'fqn'),
        corresponding to the type of the expression passed in """

    def renameByCoordinates(self, filename_path, line, col, newname):
        """ an ide friendly method which renames a class/fn/method
        pointed to by the coords and filename"""

    def extractMethod(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        """ extracts the region into the named method"""

    def extractFunction(self, filename_path, 
                        begin_x, begin_y, 
                        end_x, end_y, 
                        functionname):
        """ extracts the region into the named function"""

    def extract(self, filename_path, 
                begin_line, begin_col,
                end_line, end_col, 
                name):
        """ extracts the region into the named method/function based on context"""

    # must be an object with a write method
    def setProgressLogger(self,logger):
        pass
        
    def undo(self):
        """ undoes the last refactoring. WARNING: this is dangerous if
        the user has modified files since the last refactoring.
        Raises UndoStackEmptyException"""

    def findReferencesByCoordinates(self, filename_path, line, column):
        """ given the coords of a function, class, method or variable
        returns a generator which finds references to it.
        """

    def findDefinitionByCoordinates(self,filename_path,line,col):        
        """ given the coordates to a reference, tries to find the
        definition of that reference """
        
class NotAPythonModuleOrPackageException: pass
class CouldntLocateASTNodeFromCoordinatesException: pass

class BRMContext_impl(BRMContext):
    
    def __init__(self):
        self.ast = None
        # Used because some refactorings delegate back to the user.
        # this flag ensures that code isnt imported during those times
        self.readyToLoadNewCode = 1 
        self.paths = []
        getUndoStack(1)  # force new undo stack
        self.promptUserClientCallback = None

    def load(self, path):
        if not self.readyToLoadNewCode:
            return
            
        if(not path.endswith(".py")) and os.path.isfile(path):
            raise NotAPythonModuleOrPackageException
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        self.paths.append(path)
        class SourceVisitor:
            def visitSource(self, node):
                if node.filename == path:
                    # backup the code
                    getUndoStack().addSource(node,node.getSource())
                    getUndoStack().commitUndoFrame()
        if self.ast:
            compiler.walk(self.ast, SourceVisitor())
        self.ast = loadast(path, self.ast)


    def isLoaded(self, filename):
        if not filename.endswith(".py") or not os.path.isfile(filename):
            return 0
        if not self.ast:
            return 0
        filename = os.path.expanduser(filename)
        filename = os.path.abspath(filename)
        class SourceVisitor:
            def __init__(self):
                self.result = 0
            def visitSource(self, node):
                if node.filename == filename:
                    self.result = 1
        v = SourceVisitor()
        compiler.walk(self.ast, v)
        return v.result


    def _getAST(self):
        return self.ast

    # returns a list of saved filenames
    def save(self):
        savedfiles = saveast(self.ast)
        return savedfiles

    def setRenameMethodPromptCallback(self, callback):
        self.promptUserClientCallback = callback


    def extractMethod(self, filename_path, 
                        begin_line, begin_column, 
                        end_line, end_column, 
                        methodname):
        self.extract(filename_path, begin_line, begin_column, end_line, end_column,methodname)

    def extractFunction(self, filename_path, 
                        begin_line, begin_column, 
                        end_line, end_column, 
                        methodname):
        self.extract(filename_path, begin_line, begin_column, end_line, end_column,methodname)

    # does it based on context
    def extract(self, filename_path, 
                begin_line, begin_col,
                end_line, end_col, 
                name):
        self.refreshASTFromFileSystem()
        ast = self._getSourceNode(filename_path)
        extractMethod.extractMethod(ast, coords(begin_line, begin_col), 
                                      coords(end_line, end_col), name)

    def undo(self):
        getUndoStack().undo()

    def _promptUser(self, filename, lineno, colbegin, colend):
        return self.promptUserClientCallback(filename, lineno, colbegin, colend)


    # must be an object with a write method
    def setProgressLogger(self,logger):
        log.progress = logger


    # filename_path must be absolute
    def renameByCoordinates(self, filename_path, line, col, newname):
        self.refreshASTFromFileSystem()
        self.readyToLoadNewCode = 0
        try:
            sourcenode = self._getSourceNode(filename_path)
            rename(sourcenode,line,col,newname,self.promptUserClientCallback)
        finally:
            self.readyToLoadNewCode = 1


    def getFullyQualifiedNameOfScope(self,filename_path,line):
        self.refreshASTFromFileSystem()
        sourcenode = self._getSourceNode(filename_path)
        assert sourcenode
        scopeobj = getScopeForLine(sourcenode,line)
        return scopeobj.fqn

    def getTypeOfExpression(self, filename_path, lineno,
                           colbegin, colend):
        self.refreshASTFromFileSystem()
        sourcenode = self._getSourceNode(filename_path)
        assert sourcenode
        scopeobj = getScopeForLine(sourcenode,lineno)
        assert scopeobj
        line = sourcenode.getLines()[lineno-1]
        colbegin,colend = self._reverseCoordsIfWrongWayRound(colbegin,colend)
        src = line[colbegin:colend]
        try:
            ast = compiler.parse(src,"eval")
        except ParserError:
            raise ParserError,"Could not parse string:"+src
        type = getTypeOfExpr(scopeobj,ast.node)
        if type == UnfoundType:
            return None
        return type

    def _reverseCoordsIfWrongWayRound(self, colbegin, colend):
        if(colbegin > colend):
            colbegin,colend = colend,colbegin
        return colbegin,colend


    def findDefinitionByCoordinates(self,filename_path,line,col):
        self.refreshASTFromFileSystem()
        mod = self._getSourceNode(filename_path).fastparseroot
        return findAllPossibleDefinitionsByCoords(mod,line,col)

        
    # filename_path must be absolute
    def findReferencesByCoordinates(self, filename_path, line, column):
        self.refreshASTFromFileSystem()
        sourcenode = self._getSourceNode(filename_path)
        return findReferences(sourcenode,line,column)
        

    def _getSourceNode(self, filename):
        return self.ast.getSourceNode(filename)

    def refreshASTFromFileSystem(self):
        for path in self.paths:
            self.ast = loadast(path, self.ast)
        
def _deducePackageOfFile(filename):
    package = ""
    dot = ""
    dir = os.path.dirname(filename)
    while dir != ""and \
          os.path.exists(os.path.join(dir, "__init__.py")):
        dir, dirname = os.path.split(dir)
        package = dirname+dot+package
        dot = "."
    return package
