from __future__ import generators
from parserutils import generateLogicalLines, maskStringsAndComments, maskStringsAndRemoveComments
import re
import compiler

TABWIDTH = 4

# used so that linenum can be an attribute
class Line(str):
    pass

class StructuralNode:
    def __init__(self):
        self.childNodes = []
        self._parent = None

    def addChild(self, node):
        self.childNodes.append(node)
        node.setParent(self)

    def setParent(self, parent):
        self._parent = parent
        #print "setParent",self,parent.fqn
        if parent.fqn == "":
            self.setFQN(self.name)
        else:
            self.setFQN(parent.fqn+"."+self.name)

    def setFQN(self, fqn):
        self.fqn = fqn
        for child in self.getChildNodes():  # fqn has changed - tell children
            child.setParent(self)

    def getParent(self):
        return self._parent

    def getChildNodes(self):
        return self.childNodes

    def getChild(self,name):
        matches = [c for c in self.getChildNodes() if c.name == name]
        if matches != []:
            return matches[0]
        

    # badly named: actually returns line numbers of import statements
    def getImportLines(self):
        try:
            return self.importlines
        except AttributeError:
            return[]

    def getLinesNotIncludingThoseBelongingToChildScopes(self):
        srclines = self.module.getMaskedLines()
        lines = []
        lineno = self.getStartLine()
        for child in self.getChildNodes():
            lines+=srclines[lineno-1: child.getStartLine()-1]
            lineno = child.getEndLine()
        lines+=srclines[lineno-1: self.getEndLine()-1]
        return lines


    def generateLinesNotIncludingThoseBelongingToChildScopes(self):
        srclines = self.module.getMaskedLines()
        lines = []
        lineno = self.getStartLine()
        for child in self.getChildNodes():
            for line in srclines[lineno-1: child.getStartLine()-1]:
                yield self.attachLinenum(line,lineno)
                lineno +=1
            lineno = child.getEndLine()
        for line in srclines[lineno-1: self.getEndLine()-1]:
            yield self.attachLinenum(line,lineno)
            lineno +=1

    def attachLinenum(self,line,lineno):
        line = Line(line)
        line.linenum = lineno
        return line

class Module(StructuralNode):
    def __init__(self, name, maskedsrc):
        StructuralNode.__init__(self)
        self.name = name
        self.fqn = name  # overriden later if added to a tree
        self.indent = -TABWIDTH
        self.flattenedNodes = []
        self.maskedsrc = maskedsrc
        self._physicallines = None
        self.module = self


    def getMaskedLines(self):
        if self._physicallines is None:
            # make sure src is actually masked
            # (could just have keywords masked)
            self.maskedsrc = maskStringsAndComments(self.maskedsrc)
            self._physicallines = self.maskedsrc.splitlines(1)
        return self._physicallines

    def getFlattenedListOfChildNodes(self):
        return self.flattenedNodes

    def getStartLine(self):
        return 1

    def getEndLine(self):
        return len(self.getMaskedLines())+1


    def getSourceNode(self):
        return self.sourcenode

    def setSourceNode(self, sourcenode):
        self.sourcenode = sourcenode

    def matchesCompilerNode(self,node):
        return isinstance(node,compiler.ast.Module) and \
               node.name == self.name

indentRE = re.compile("^(\s*)\S")
class Node:
    # module = the module node
    # linenum = starting line number
    def __init__(self, name, module, linenum, indent):
        self.name = name
        self.module = module
        self.linenum = linenum
        self.endline = None
        self.indent = indent

    def getMaskedLines(self):
        return self.module.getMaskedLines()[self.getStartLine()-1:self.getEndLine()-1]

    def getStartLine(self):
        return self.linenum

    def getEndLine(self):
        if self.endline is None:
            physicallines = self.module.getMaskedLines()
            lineno = self.linenum
            logicallines = generateLogicalLines(physicallines[lineno-1:])

            # skip the first line, because it's the declaration
            line = logicallines.next()
            lineno+=line.count("\n")

            # scan to the end of the fn
            for line in logicallines:
                #print lineno,":",line,
                match = indentRE.match(line)
                if match and match.end()-1 <= self.indent:
                    break
                lineno+=line.count("\n")
            self.endline = lineno
        return self.endline

    # linenum starts at 0
    def getLine(self, linenum):
        return self.module.getSourceNode().getLines()[(self.getStartLine()-1) + linenum]


baseClassesRE = re.compile("class\s+[^(]+\((.+)\):")

classNameRE = re.compile("^\s*class (\w+)")

class Class(StructuralNode, Node):
    def __init__(self, name, module, linenum, indent):
        StructuralNode.__init__(self)
        Node.__init__(self, name, module, linenum, indent)
        self.type = "Class"

    def getBaseClassNames(self):
        match = baseClassesRE.search(self.getLine(0))
        if match:
            return [s.strip()for s in match.group(1).split(",")]
        else:
            return []

    def getColumnOfName(self):
        match = classNameRE.match(self.getLine(0))
        return match.start(1)

    def __repr__(self):
        return "<bike:Class:%s>" % self.name

    def __str__(self):
        return "bike:Class:"+self.name

    def matchesCompilerNode(self,node):
        return isinstance(node,compiler.ast.Class) and \
               node.name == self.name



fnNameRE = re.compile("^\s*def (\w+)")

class Function(StructuralNode, Node):
    def __init__(self, name, module, linenum, indent):
        StructuralNode.__init__(self)
        Node.__init__(self, name, module, linenum, indent)
        self.type = "Function"

    def getColumnOfName(self):
        match = fnNameRE.match(self.getLine(0))
        return match.start(1)

    def __repr__(self):
        return "<bike:Function:%s>" % self.name

    def __str__(self):
        return "bike:Function:"+self.name

    def matchesCompilerNode(self,node):
        return isinstance(node,compiler.ast.Function) and \
               node.name == self.name
